/*
 * Copyright 2012 Peter W Leong
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package biz.freshcode.b_generation;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import static biz.freshcode.b_generation.BeanBuilderGenerator.BUILDER_SUFFIX;
import static biz.freshcode.b_generation.BgenStringUtil.isBlank;

/**
 * Writes classes to a defined output folder.
 */
public class BeanBuilderWriter {
    private BeanBuilderGenerator generator = new DefaultBeanBuilderGenerator();
    private File outputFolder = new File("src" + File.separator + "main" + File.separator + "java");
    private PackageMapper mapper = new PackageMapper();

    /**
     * Assign the desired generator (defaults to DefaultBeanBuilderGenerator).
     */
    public BeanBuilderWriter generator(BeanBuilderGenerator g) {
        generator = g;
        return this;
    }

    /**
     * Sets up a mapping from one package prefix to another.
     * These are traversed in the same order added.  Add a "" mapping to catch all.
     */
    public BeanBuilderWriter addMapping(String from, String to) {
        mapper.addMapping(from, to);
        return this;
    }

    /**
     * Sets up the output folder for generated sources (defaults to 'src/main/java').
     */
    public BeanBuilderWriter outputFolder(File f) {
        outputFolder = f;
        return this;
    }

    /**
     * Outputs an XxxBuilder.java file for each given bean class.
     */
    public void write(Class... beanClasses) {
        for (Class cls : beanClasses) {
            String pkg = mapper.calcPkgFor(cls);
            String path = pkg.replaceAll("\\.", File.separator);
            String src = generator.generate(cls, pkg);
            if (!isBlank(path)) path += File.separator;
            path += cls.getSimpleName() + BUILDER_SUFFIX + ".java";
            File output = new File(outputFolder, path);
            writeFile(src, output);
        }
    }

    private void writeFile(String contents, File output) {
        // create any necessary folders
        File srcFolder = output.getParentFile();
        if (!srcFolder.exists() && !srcFolder.mkdirs()) {
            throw new IllegalStateException("Unable to mkdirs for " + srcFolder);
        }

        FileWriter fw = null;
        try {
            fw = new FileWriter(output);
            fw.write(contents);
        } catch (IOException e) {
            throw new RuntimeException("Unable to write " + output, e);
        } finally {
            try {
                if (fw != null) fw.close();
            } catch (IOException e) {
                // nothing
            }
        }
    }
}
