/*
 * Copyright 2012 Peter W Leong
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package biz.freshcode.b_generation;

import java.lang.reflect.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import static biz.freshcode.b_generation.BgenStringUtil.isBlank;
import static biz.freshcode.b_generation.BgenStringUtil.join;
import static java.util.Arrays.asList;

/**
 * Utility methods.
 */
public class BgenUtil {

    public static final Package JAVA_LANG = String.class.getPackage();
    private static final String DOTx3 = "...";
    private static final String REGEX_DOLLAR = "\\$";


    /**
     * Returns a argument declaration list for the given method.  EG: String v0, int v1.
     *
     * @see #genericStringToArgDecl(String, boolean, boolean)
     */
    public static String argDecl(Method m, boolean forJavadoc) {
        return genericStringToArgDecl(m.toGenericString(), m.isVarArgs(), forJavadoc);
    }

    /**
     * Overloaded variant.
     *
     * @see #argDecl(java.lang.reflect.Method, boolean)
     */
    public static String argDecl(Constructor c, boolean forJavadoc) {
        return genericStringToArgDecl(c.toGenericString(), c.isVarArgs(), forJavadoc);
    }

    /**
     * Returns a wouldbe arg declaration for a method or constructor.
     *
     * @param genericStr From Constructor.toGenericString() or Method.toGenericString()
     * @param hasVarArgs True if the last arg should have '...' and not []
     * @param forJavadoc False if variable names v0, v1... will be included
     * @return Arg declaration like "String v0, Integer v1, int... v2"
     */
    public static String genericStringToArgDecl(String genericStr, boolean hasVarArgs, boolean forJavadoc) {
        int ixLb = genericStr.indexOf("(");
        int ixRb = genericStr.lastIndexOf(")");
        String methodDecl = genericStr.substring(ixLb + 1, ixRb);
        if (isBlank(methodDecl)) return "";

        List<String> l = splitTypes(methodDecl);

        repairNestedDollars(l);

        int count = l.size();

        for (int i = 0; i < count; i++) {
            String si = l.get(i).trim();
            if (hasVarArgs &&
                    i == count - 1 && // last arg
                    !si.contains(DOTx3)) { // some JDKs don't need this
                // process var args variation
                int ixVarArg = si.lastIndexOf("[]");
                si = si.substring(0, ixVarArg) + DOTx3;
            }

            if (!forJavadoc) si = si + " v" + i; // add var names
            else {
                // no angle brackets for javadoc
                int ixLt = si.indexOf('<');
                if (ixLt >= 0) si = si.substring(0, ixLt);
            }

            l.set(i, si);
        }

        return join(l, ", ");
    }

    /**
     * Intelligently splits a parameter types signature into types.
     * EG: 'String, Map<T, U>' => 'String', 'Map<T, U>'
     */
    public static List<String> splitTypes(String methodDecl) {
        List<String> l = asList(methodDecl.split(","));
        List<String> result = new ArrayList<String>();
        String prefix = "";
        for (String s : l) {
            s = s.trim();
            s = prefix + s;
            prefix = "";

            if (isAngleBalanced(s)) result.add(s);
            else prefix = s + ", ";
        }
        if (!isBlank(prefix)) {
            throw new IllegalArgumentException("'" + methodDecl + "' cannot be intelligently split.");
        }
        return result;
    }

    /**
     * Replace pkg.Host$Nested with pkg.Host.Nested
     * NOTE: some JVM's repeat the entire package like 'pkg.Host.pkg.Host$Nested'.
     */
    private static String repairNestedDollar(String cls) {
        int ixDollar = cls.lastIndexOf("$");
        if (ixDollar < 0) return cls; // nothing to do

        String p0 = cls.substring(0, ixDollar);
        String p1 = cls.substring(ixDollar + 1);
        int len = p0.length();
        if (len > 3) {
            if (len % 2 == 1) {
                // odd length
                int ixDot = len / 2;
                if (p0.charAt(ixDot) == '.') {
                    // dot in the middle
                    String clsHost = p0.substring(0, ixDot);
                    if (p0.endsWith(clsHost)) {
                        // package repeated
                        p0 = clsHost;
                    }
                }
            }
        }
        return repairNestedDollar(p0) + "." + p1;
    }

    private static void repairNestedDollars(List<String> types) {
        for (int i = 0; i < types.size(); i++) {
            String t = types.get(i);
            if (t.contains("$")) {
                // nested type
                int ix = t.indexOf("<");


                String s1 = t;
                String suffix = "";
                if (ix >= 0) {
                    // There are type args
                    s1 = t.substring(0, ix);
                    suffix = t.substring(ix);
                }
                String altT = repairNestedDollar(s1) + suffix;
                types.set(i, altT);
            }
        }
    }

    private static boolean isAngleBalanced(String s) {
        int ltCount = countOccurences(s, '<');
        int gtCount = countOccurences(s, '>');
        return ltCount == gtCount;
    }

    private static int countOccurences(String s, char ch) {
        int count = 0;
        for (char c : s.toCharArray()) {
            if (c == ch) count++;
        }
        return count;
    }

    /**
     * Returns the generic type description to be used for Variables.  EG: &lt;T,U>
     */
    public static String genericType(TypeVariable[] tvs) {
        String s = "";
        for (TypeVariable tv : tvs) {
            s = join(s, ", ", tv.getName());
        }
        return angles(s);
    }

    /**
     * Reproduces the class generic declaration.  EG: &lt;T extends Comparable, U>
     */
    public static String classDecl(TypeVariable[] tps) {
        return classDecl(tps, new TypeVariable[0]);
    }

    /**
     * Same as simpler version except TypeVariables in the 'known' array are not expanded.
     * This is useful for reproducing Type params for a method where class level type params are not expanded.
     */
    public static String classDecl(TypeVariable[] tps, TypeVariable[] known) {
        List<TypeVariable> list = new ArrayList<TypeVariable>();
        Collections.addAll(list, known);
        return genericDecl(tps, list);
    }

    /**
     * Returns the fully qualified class name for the given class, handline nested classes as appropriate.
     */
    // TODO: Look at Class.getSimpleName() for weird cases.
    public static String className(Class beanCls) {
        return beanCls.getName().replaceAll(REGEX_DOLLAR, ".");
    }

    private static String decl(Type t, List<TypeVariable> known) {
        if (t instanceof Class) return classDecl((Class) t);
        else if (t instanceof TypeVariable) return typeVarDecl((TypeVariable) t, known);
        else if (t instanceof ParameterizedType) return paramTypeDecl((ParameterizedType) t, known);
        else return "";
    }

    private static String classDecl(Class t) {
        String name = t.getName();
        if (JAVA_LANG.equals(t.getPackage())) {
            String JAVA_LANG_STR = JAVA_LANG.getName();
            name = name.substring(JAVA_LANG_STR.length() + 1);
        }
        return name;
    }

    private static String genericDecl(Type[] tps, List<TypeVariable> known) {
        String s = "";
        for (Type tv : tps) {
            s = join(s, ", ", decl(tv, known));
        }
        return angles(s);
    }

    private static String typeVarDecl(TypeVariable tv, List<TypeVariable> known) {
        String str = tv.getName();
        Type[] bounds = tv.getBounds();

        // only default bounds
        if (isDefaultBounds(bounds)) return str;

        // not expanding
        if (known.contains(tv)) return str;

        // prevent stack overflow
        known.add(tv);

        String ext = "";

        for (Type t : bounds) {
            String st = decl(t, known);
            ext = join(ext, " & ", st);
        }
        return join(str, " extends ", ext);
    }

    private static boolean isDefaultBounds(Type[] bounds) {
        return bounds.length == 1 && bounds[0].equals(Object.class);
    }

    private static String paramTypeDecl(ParameterizedType t, List<TypeVariable> known) {
        return decl(t.getRawType(), known) + genericDecl(t.getActualTypeArguments(), known);
    }

    private static String angles(String s) {
        if (isBlank(s)) return s;
        return "<" + s + ">";
    }
}
