/*
 * Copyright 2012 Peter W Leong
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package biz.freshcode.b_generation;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.TypeVariable;
import java.util.Arrays;
import java.util.Comparator;

import static biz.freshcode.b_generation.BgenStringUtil.*;
import static biz.freshcode.b_generation.BgenUtil.argDecl;

/**
 * Generates a bean builder so instead of:
 * <pre>
 * Margins m = new Margins();
 * m.setTop(0);
 * m.setRight(5);
 * m.setBottom(0);
 * m.setLeft(5);
 * return m;
 * </pre>
 * we can use
 * <pre>
 * return new MarginsBuilder()
 * .top(0)
 * .right(5)
 * .bottom(0)
 * .left(5)
 * .margins;
 *
 * </pre>
 * <p/>
 * Usage:
 * <pre>
 *   DefaultBeanBuilderGenerator gen = new DefaultBeanBuilderGenerator();
 *   System.out.println(gen.generate(BgenDemo.class)); //<<<<<<<<<<<<<<<<<< The bean class here
 * </pre>
 */
public class DefaultBeanBuilderGenerator implements BeanBuilderGenerator {
    public static final String SPC = " ";
    public static final String NL = "\n";
    /*
    * Ideas:
    * - Regeneration capabilities
    * - Annotation indicating a custom method that should be retained
    * - Protected methods for customising behaviour
    */
    private final String IN;
    private final String INx2;

    {
        IN = getIndent();
        INx2 = IN + IN;
    }

    public String generate(Class beanCls) {
        String packageName = beanCls.getPackage().getName();
        return generate(beanCls, packageName);
    }

    public String generate(Class beanCls, String pkgName) {
        Info o = new Info(beanCls);
        String src = "package " + pkgName + ";";

        if (!beanCls.getPackage().getName().equals(pkgName)) {
            src += NL + NL + "import " + BgenUtil.className(beanCls) + ";";
        }

        src += NL;
        src += javadoc("Generated by " + getClass().getSimpleName(), o.pkg_Xxx, 0);
        src = join(src, NL, getBuilderAnnotation(beanCls));
        src += NL + "public class " + o.XxxBuilder + BgenUtil.classDecl(o.typeParams);
        src = join(src, " extends ", getBuilderParent(beanCls, o.XxxBuilderT));
        src += " {";

        src += line(IN + "public final " + o.XxxT + SPC + o.xxx);

        src += NL + NL + IN + "public " + o.XxxBuilder + "(" + o.XxxT + " v) {";
        src += line(INx2 + o.xxx + " = v");
        src += NL + IN + "}";

        for (Constructor ctor : beanCls.getConstructors()) {
            boolean isPublic = Modifier.isPublic(ctor.getModifiers());
            if (isPublic && !ctor.isSynthetic()) {
                src += genContructor(ctor, o);
            }
        }

        for (Method m : sortedMethods(beanCls)) {
            if (isSetter(m) || isAdder(m)) {
                src += genMethod(m, o);
            }
        }

        src += NL + "}" + NL;
        return src;
    }

    /**
     * Returns the parent class designation if any or an empty string if there is not parent class.
     */
    protected String getBuilderParent(Class beanCls, String builderType) {
        return "";
    }

    /**
     * The string to use for indent of code.  Defaults to 4x SPACE.
     */
    protected String getIndent() {
        return "    ";
    }

    private String javadoc(String msg, String see, int indentCount) {
        String linePrefix = "";
        for (int i = 0; i < indentCount; i++) linePrefix += IN;
        linePrefix = NL + linePrefix;
        String s;
        s = linePrefix + "/**";
        if (!isBlank(msg)) {
            s += linePrefix + " * " + msg;
            s += linePrefix + " *";
        }
        s += linePrefix + " * @see " + see;
        s += linePrefix + " */";
        return s;
    }

    private Method[] sortedMethods(Class beanCls) {
        Method[] methods = beanCls.getMethods();
        Arrays.sort(methods, new Comparator<Method>() {
            public int compare(Method o1, Method o2) {
                int byName = o1.getName().compareTo(o2.getName());
                if (byName == 0) return o1.getParameterTypes().length - o2.getParameterTypes().length;
                return byName;
            }
        });
        return methods;
    }

    private String genMethod(Method m, Info o) {
        String mstr = NL;
        mstr += javadocMethod(o, m);
        mstr += NL + IN + "public";

        String methodTypeParams = BgenUtil.classDecl(m.getTypeParameters(), o.typeParams);
        mstr = join(mstr, SPC, methodTypeParams);

        String methodName = isSetter(m) ? shortName(m) : m.getName();
        mstr += SPC + o.XxxBuilderT + SPC + methodName + "(" + argDecl(m, false) + ") {";
        int argCount = m.getParameterTypes().length;
        mstr += line(INx2 + o.xxx + "." + m.getName() + "(" + methodArgs(argCount) + ")");
        mstr += line(INx2 + "return this");
        mstr += NL + IN + "}";
        return mstr;
    }

    private String genContructor(Constructor ctor, Info o) {
        String cstr = NL;
        cstr += javadocCtor(o, ctor);
        cstr += NL + IN + "public ";
        cstr += o.XxxBuilder + "(" + argDecl(ctor, false) + ") {";
        int argCount = ctor.getParameterTypes().length;
        cstr += line(INx2 + "this(new " + o.XxxT + "(" + methodArgs(argCount) + "))");
        cstr += NL + IN + "}";
        return cstr;
    }

    private String javadocCtor(Info o, Constructor ctor) {
        // TODO: Sort out default constructor issue... target of @see is not there.
        return javadocMethod(o, o.Xxx, argDecl(ctor, true));
    }

    private String javadocMethod(Info o, Method m) {
        String methodName = m.getName();
        return javadocMethod(o, methodName, argDecl(m, true));
    }

    private String javadocMethod(Info o, String methodName, String signature) {
        return javadoc("", o.pkg_Xxx + "#" + methodName + "(" + signature + ")", 1);
    }

    // build up "v0, v1, v2..."
    private String methodArgs(int paramCount) {
        String args = "";
        for (int i = 0; i < paramCount; i++) {
            args = join(args, ", ", "v" + i);
        }
        return args;
    }

    private static boolean isSetter(Method m) {
        String name = m.getName();
        return name.length() > 3
                && name.startsWith("set")
                && Character.isUpperCase(name.charAt(3))
                && eligible(m);
    }

    private static boolean eligible(Method m) {
        int mods = m.getModifiers();
        return !Modifier.isPrivate(mods)
                && !Modifier.isStatic(mods)
                && !m.isSynthetic()
                && isTrivialReturnType(m);
    }

    /**
     * Indicates the return type can be ignored.
     * Chaining methods are considered simple enough to include in builders.  EG: GWT's MenuBar.addItem()
     */
    private static boolean isTrivialReturnType(Method m) {
        Class<?> returnType = m.getReturnType();
        boolean isVoid = Void.TYPE.equals(returnType);
        Class<?>[] paramTypes = m.getParameterTypes();
        boolean isChained = (paramTypes.length == 1) && (returnType.equals(paramTypes[0]) || returnType.equals(m.getDeclaringClass()));
        return isVoid || isChained;
    }

    private static boolean isAdder(Method m) {
        String name = m.getName();
        return name.startsWith("add")
                && (name.length() == 3 || Character.isUpperCase(name.charAt(3)))
                && eligible(m);
    }

    /**
     * Will return 'bruce' from method 'setBruce'.
     */
    private static String shortName(Method setter) {
        String setLess = setter.getName().substring(3);
        return decapitalize(setLess);
    }

    private static String line(String contents) {
        return NL + contents + ";";
    }

    /**
     * The annotation to place at top of each builder class (if any).
     */
    protected String getBuilderAnnotation(@SuppressWarnings("UnusedParameters") Class beanClass) {
        return "";
    }

    class Info {
        final String Xxx;
        final String pkg_Xxx;
        final TypeVariable[] typeParams;
        final String xxx;
        final String XxxT;
        final String XxxBuilder;
        final String XxxBuilderT;

        Info(Class cls) {
            Xxx = cls.getSimpleName();
            pkg_Xxx = BgenUtil.className(cls);
            XxxBuilder = Xxx + BUILDER_SUFFIX;
            xxx = decapitalize(Xxx);
            typeParams = cls.getTypeParameters();
            String genericType = BgenUtil.genericType(typeParams);
            XxxT = Xxx + genericType;
            XxxBuilderT = XxxBuilder + genericType;
        }
    }
}
